@@ -1,3 +1,11 @@
+"""
+The implementation of the `isqlite` command-line tool.
+
+A subcommand of `isqlite` called `xyz` will be implemented in a function called
+`main_xyz`, with a corresponding wrapper function called `main_xyz_wrapper` that is
+used by Click, the argument parsing framework. The tests in `tests/test_cli.py` call
+`main_xyz` directly.
+"""
 import collections
 import importlib
 import readline  # noqa: F401
@@ -32,6 +40,7 @@ DESC_HELP = (
     "When combined with --order-by, order the results in descending rather than "
     + "ascending order."
 )
+NO_CONFIRM_HELP = "Do not prompt for confirmation."
 
 
 @click.group()
@@ -164,10 +173,20 @@ def main_create_table(db_path, table, columns):
 @click.argument("table")
 @click.argument("pk", type=int, required=False, default=None)
 @click.option("-w", "--where", default="")
-def main_delete(db_path, table, pk, *, where):
+@click.option(
+    "--no-confirm",
+    is_flag=True,
+    default=False,
+    help=NO_CONFIRM_HELP,
+)
+def main_delete_wrapper(*args, **kwargs):
     """
     Delete a row.
     """
+    main_delete(*args, **kwargs)
+
+
+def main_delete(db_path, table, pk=None, *, where="", no_confirm=False):
     with Database(db_path) as db:
         if pk is not None:
             try:
@@ -186,26 +205,28 @@ def main_delete(db_path, table, pk, *, where):
 
                 prettyprint_row(row)
 
-            print()
-            if not click.confirm("Are you sure you wish to delete this record?"):
+            if not no_confirm:
                 print()
-                print("Operation aborted.")
-                sys.exit(1)
+                if not click.confirm("Are you sure you wish to delete this record?"):
+                    print()
+                    print("Operation aborted.")
+                    sys.exit(1)
 
             db.delete_by_rowid(table, pk)
             print()
             print(f"Row {pk} deleted from table {table!r}.")
         else:
-            n = db.count(table, where=where)
-            if not where:
-                msg = f"Are you sure you wish to delete ALL {n} row(s) from {table!r}?"
-            else:
-                msg = f"Are you sure you wish to delete {n} row(s) from {table!r}?"
+            if not no_confirm:
+                n = db.count(table, where=where)
+                if not where:
+                    msg = f"Are you sure you wish to delete ALL {n} row(s) from {table!r}?"
+                else:
+                    msg = f"Are you sure you wish to delete {n} row(s) from {table!r}?"
 
-            if not click.confirm(msg):
-                print()
-                print("Operation aborted.")
-                sys.exit(1)
+                if not click.confirm(msg):
+                    print()
+                    print("Operation aborted.")
+                    sys.exit(1)
 
             # `Database.delete` doesn't accept a blank `where` parameter for safety
             # reasons, so we use an explicit WHERE clause that will match every row.
@@ -219,46 +240,69 @@ def main_delete(db_path, table, pk, *, where):
 
 @cli.command(name="drop-column")
 @click.option("--db", "db_path", envvar="ISQLITE_DB")
+@click.option("--schema", "schema_path", envvar="ISQLITE_SCHEMA")
 @click.argument("table")
 @click.argument("column")
-def main_drop_column(db_path, table, column):
+@click.option(
+    "--no-confirm",
+    is_flag=True,
+    default=False,
+    help=NO_CONFIRM_HELP,
+)
+def main_drop_column_wrapper(*args, **kwargs):
     """
     Drop a column from a table.
     """
-    with Database(db_path) as db:
-        count = db.count(table)
-        print(f"WARNING: Table {table!r} contains {count} row(s) of data.")
-        print()
-        if not click.confirm("Are you sure you wish to drop this column?"):
-            print()
-            print("Operation aborted.")
-            sys.exit(1)
+    main_drop_column(*args, **kwargs)
 
-        db.drop_column(table, column)
+
+def main_drop_column(db_path, schema_path, table, column, *, no_confirm=False):
+    schema_module = get_schema_module(schema_path)
+    if not no_confirm:
+        with Database(db_path) as db:
+            count = db.count(table)
+
+            print(f"WARNING: Table {table!r} contains {count} row(s) of data.")
+            print()
+            if not click.confirm("Are you sure you wish to drop this column?"):
+                print()
+                print("Operation aborted.")
+                sys.exit(1)
+
+    with DatabaseMigrator(db_path, schema_module=schema_module) as migrator:
+        migrator.drop_column(table, column)
         print()
         print(f"Column {column!r} dropped from table {table!r}.")
 
 
 @cli.command(name="drop-table")
 @click.option("--db", "db_path", envvar="ISQLITE_DB")
-@click.option("--schema", "schema_path", envvar="ISQLITE_SCHEMA")
 @click.argument("table")
-def main_drop_table(db_path, schema_path, table):
+@click.option(
+    "--no-confirm",
+    is_flag=True,
+    default=False,
+    help=NO_CONFIRM_HELP,
+)
+def main_drop_table_wrapper(*args, **kwargs):
     """
     Drop a table from the database.
     """
-    with Database() as db:
-        count = db.count(table)
-        print(f"WARNING: Table {table!r} contains {count} row(s) of data.")
-        print()
-        if not click.confirm("Are you sure you wish to drop this table?"):
-            print()
-            print("Operation aborted.")
-            sys.exit(1)
+    main_drop_table(*args, **kwargs)
 
-    schema_module = get_schema_module(schema_path)
-    with DatabaseMigrator(schema_module=schema_module) as migrator:
-        migrator.drop_table(table)
+
+def main_drop_table(db_path, table, *, no_confirm=False):
+    with Database(db_path) as db:
+        if not no_confirm:
+            count = db.count(table)
+            print(f"WARNING: Table {table!r} contains {count} row(s) of data.")
+            print()
+            if not click.confirm("Are you sure you wish to drop this table?"):
+                print()
+                print("Operation aborted.")
+                sys.exit(1)
+
+        db.drop_table(table)
         print()
         print(f"Table {table!r} dropped from the database.")
 
@@ -268,14 +312,22 @@ def main_drop_table(db_path, schema_path, table):
 @click.option("--schema", "schema_path", envvar="ISQLITE_SCHEMA")
 @click.argument("table")
 @click.argument("pk", type=int)
-def main_get(db_path, schema_path, table, pk):
+def main_get_wrapper(*args, **kwargs):
     """
     Fetch a single row.
     """
-    schema_module = get_schema_module(schema_path)
-    schema_dict = schema_module_to_dict(schema_module)
+    main_get(*args, **kwargs)
+
+
+def main_get(db_path, schema_path, table, pk):
+    if schema_path:
+        schema_module = get_schema_module(schema_path)
+        schema_dict = schema_module_to_dict(schema_module)
+    else:
+        schema_module = None
+
     with Database(db_path, readonly=True, schema_module=schema_module) as db:
-        row = db.get_by_rowid(table, pk, get_related=True)
+        row = db.get_by_rowid(table, pk, get_related=schema_module is not None)
         for key, value in row.items():
             if isinstance(value, collections.OrderedDict):
                 row[key] = get_column_as_string(schema_dict, table, key, value)
@@ -427,6 +479,35 @@ def base_list(
             prettyprint_rows(rows, columns=columns, hide=hide, page=page)
 
 
+@cli.command(name="list-tables")
+@click.option("--db", "db_path", envvar="ISQLITE_DB")
+def main_list_tables_wrapper(*args, **kwargs):
+    """
+    List the names of the tables in the database.
+    """
+    main_list_tables(*args, **kwargs)
+
+
+def main_list_tables(db_path):
+    with Database(db_path, readonly=True) as db:
+        rows = db.sql(
+            """
+            SELECT
+              name
+            FROM
+              sqlite_master
+            WHERE
+              type = 'table'
+            AND
+              name NOT LIKE 'sqlite_%'
+            """,
+            as_tuple=True,
+        )
+
+        if rows:
+            print("\n".join(row[0] for row in rows))
+
+
 @cli.command(name="migrate")
 @click.option("--db", "db_path", envvar="ISQLITE_DB")
 @click.option("--schema", "schema_path", envvar="ISQLITE_SCHEMA")
@@ -550,10 +631,14 @@ def main_migrate(db_path, schema_path, table, *, write, no_backup, debug):
 @click.argument("table")
 @click.argument("old_name")
 @click.argument("new_name")
-def main_rename_column(schema_path, db_path, table, old_name, new_name):
+def main_rename_column_wrapper(*args, **kwargs):
     """
     Rename a column.
     """
+    main_rename_column(*args, **kwargs)
+
+
+def main_rename_column(db_path, schema_path, table, old_name, new_name):
     schema_module = get_schema_module(schema_path)
     with DatabaseMigrator(db_path, schema_module=schema_module) as migrator:
         migrator.rename_column(table, old_name, new_name)
@@ -562,16 +647,18 @@ def main_rename_column(schema_path, db_path, table, old_name, new_name):
 
 @cli.command(name="rename-table")
 @click.option("--db", "db_path", envvar="ISQLITE_DB")
-@click.option("--schema", "schema_path", envvar="ISQLITE_SCHEMA")
 @click.argument("table")
 @click.argument("new_name")
-def main_rename_table(schema_path, db_path, table, new_name):
+def main_rename_table_wrapper(*args, **kwargs):
     """
     Rename a table.
     """
-    schema_module = get_schema_module(schema_path)
-    with DatabaseMigrator(db_path, schema_module=schema_module) as migrator:
-        migrator.rename_table(table, new_name)
+    main_rename_table(*args, **kwargs)
+
+
+def main_rename_table(db_path, table, new_name):
+    with Database(db_path) as db:
+        db.rename_table(table, new_name)
         print(f"Table {table!r} renamed to {new_name!r}.")
 
 
@@ -580,10 +667,14 @@ def main_rename_table(schema_path, db_path, table, new_name):
 @click.option("--schema", "schema_path", envvar="ISQLITE_SCHEMA")
 @click.argument("table")
 @click.argument("columns", nargs=-1)
-def main_reorder_columns(schema_path, db_path, table, columns):
+def main_reorder_columns_wrapper(*args, **kwargs):
     """
     Change the order of columns in a table.
     """
+    main_reorder_columns(*args, **kwargs)
+
+
+def main_reorder_columns(db_path, schema_path, table, columns):
     schema_module = get_schema_module(schema_path)
     with DatabaseMigrator(db_path, schema_module=schema_module) as migrator:
         migrator.reorder_columns(table, columns)
@@ -603,24 +694,28 @@ def main_reorder_columns(schema_path, db_path, table, columns):
 @click.option("--offset", default=None, help=OFFSET_HELP)
 @click.option("--order-by", multiple=True, default=[], help=ORDER_BY_HELP)
 @click.option("--desc", is_flag=True, default=False, help=DESC_HELP)
+def main_search_wrapper(*args, **kwargs):
+    """
+    Shorthand for `list <table> -s <query>`
+    """
+    main_search(*args, **kwargs)
+
+
 def main_search(
     db_path,
     schema_path,
     table,
     query,
     *,
-    where,
-    columns,
-    hide,
-    page,
-    limit,
-    offset,
-    order_by,
-    desc,
+    where="",
+    columns=[],
+    hide=[],
+    page=1,
+    limit=None,
+    offset=None,
+    order_by=[],
+    desc=False,
 ):
-    """
-    Shorthand for `list <table> -s <query>`
-    """
     base_list(
         db_path,
         schema_path,
@@ -664,10 +759,43 @@ def main_sql(db_path, query, *, columns, hide, page, write):
 
 @cli.command(name="update")
 @click.option("--db", "db_path", envvar="ISQLITE_DB")
+@click.argument("table")
+@click.argument("pk", type=int)
+@click.argument("payload", nargs=-1)
+@click.option(
+    "--auto-timestamp",
+    is_flag=True,
+    default=True,
+    help="Automatically populate `last_updated_at` column with current time.",
+)
+def main_update_wrapper(*args, **kwargs):
+    """
+    Update an existing row non-interactively.
+
+    PAYLOAD should be a list of space-separated key-value pairs, e.g.
+
+        isqlite update --db db.sqlite3 my_table 123 a=1 b=2
+    """
+    main_update(*args, **kwargs)
+
+
+def main_update(db_path, table, pk, payload, *, auto_timestamp=True):
+    payload_as_map = {}
+    for key_value in payload:
+        key, value = key_value.split("=")
+        payload_as_map[key] = value
+
+    with Database(db_path) as db:
+        db.update_by_rowid(table, pk, payload_as_map, auto_timestamp=auto_timestamp)
+        print(f"Row {pk} updated.")
+
+
+@cli.command(name="iupdate")
+@click.option("--db", "db_path", envvar="ISQLITE_DB")
 @click.option("--schema", "schema_path", envvar="ISQLITE_SCHEMA")
 @click.argument("table")
 @click.argument("pk", type=int)
-def main_update(db_path, schema_path, table, pk):
+def main_iupdate(db_path, schema_path, table, pk):
     """
     Update an existing row interactively.
     """
